# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: button_service.proto
# pylint: skip-file
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='button_service.proto',
  package='chip.rpc',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x14\x62utton_service.proto\x12\x08\x63hip.rpc\"\r\n\x0b\x42uttonEmpty\"*\n\x0b\x42uttonEvent\x12\x0b\n\x03idx\x18\x01 \x01(\r\x12\x0e\n\x06pushed\x18\x02 \x01(\x08\x32\x41\n\x06\x42utton\x12\x37\n\x05\x45vent\x12\x15.chip.rpc.ButtonEvent\x1a\x15.chip.rpc.ButtonEmpty\"\x00\x62\x06proto3'
)




_BUTTONEMPTY = _descriptor.Descriptor(
  name='ButtonEmpty',
  full_name='chip.rpc.ButtonEmpty',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=34,
  serialized_end=47,
)


_BUTTONEVENT = _descriptor.Descriptor(
  name='ButtonEvent',
  full_name='chip.rpc.ButtonEvent',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='idx', full_name='chip.rpc.ButtonEvent.idx', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='pushed', full_name='chip.rpc.ButtonEvent.pushed', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=49,
  serialized_end=91,
)

DESCRIPTOR.message_types_by_name['ButtonEmpty'] = _BUTTONEMPTY
DESCRIPTOR.message_types_by_name['ButtonEvent'] = _BUTTONEVENT
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ButtonEmpty = _reflection.GeneratedProtocolMessageType('ButtonEmpty', (_message.Message,), {
  'DESCRIPTOR' : _BUTTONEMPTY,
  '__module__' : 'button_service_pb2'
  # @@protoc_insertion_point(class_scope:chip.rpc.ButtonEmpty)
  })
_sym_db.RegisterMessage(ButtonEmpty)

ButtonEvent = _reflection.GeneratedProtocolMessageType('ButtonEvent', (_message.Message,), {
  'DESCRIPTOR' : _BUTTONEVENT,
  '__module__' : 'button_service_pb2'
  # @@protoc_insertion_point(class_scope:chip.rpc.ButtonEvent)
  })
_sym_db.RegisterMessage(ButtonEvent)



_BUTTON = _descriptor.ServiceDescriptor(
  name='Button',
  full_name='chip.rpc.Button',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=93,
  serialized_end=158,
  methods=[
  _descriptor.MethodDescriptor(
    name='Event',
    full_name='chip.rpc.Button.Event',
    index=0,
    containing_service=None,
    input_type=_BUTTONEVENT,
    output_type=_BUTTONEMPTY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_BUTTON)

DESCRIPTOR.services_by_name['Button'] = _BUTTON

# @@protoc_insertion_point(module_scope)
