# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: device_service.proto
# pylint: skip-file
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='device_service.proto',
  package='chip.rpc',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x14\x64\x65vice_service.proto\x12\x08\x63hip.rpc\"\r\n\x0b\x44\x65viceEmpty\"d\n\nDeviceInfo\x12\x11\n\tvendor_id\x18\x01 \x01(\r\x12\x12\n\nproduct_id\x18\x02 \x01(\r\x12\x18\n\x10software_version\x18\x03 \x01(\r\x12\x15\n\rserial_number\x18\x04 \x01(\t2\x80\x02\n\x06\x44\x65vice\x12>\n\x0c\x46\x61\x63toryReset\x12\x15.chip.rpc.DeviceEmpty\x1a\x15.chip.rpc.DeviceEmpty\"\x00\x12\x38\n\x06Reboot\x12\x15.chip.rpc.DeviceEmpty\x1a\x15.chip.rpc.DeviceEmpty\"\x00\x12<\n\nTriggerOta\x12\x15.chip.rpc.DeviceEmpty\x1a\x15.chip.rpc.DeviceEmpty\"\x00\x12>\n\rGetDeviceInfo\x12\x15.chip.rpc.DeviceEmpty\x1a\x14.chip.rpc.DeviceInfo\"\x00\x62\x06proto3'
)




_DEVICEEMPTY = _descriptor.Descriptor(
  name='DeviceEmpty',
  full_name='chip.rpc.DeviceEmpty',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=34,
  serialized_end=47,
)


_DEVICEINFO = _descriptor.Descriptor(
  name='DeviceInfo',
  full_name='chip.rpc.DeviceInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='vendor_id', full_name='chip.rpc.DeviceInfo.vendor_id', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='product_id', full_name='chip.rpc.DeviceInfo.product_id', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='software_version', full_name='chip.rpc.DeviceInfo.software_version', index=2,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='serial_number', full_name='chip.rpc.DeviceInfo.serial_number', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=49,
  serialized_end=149,
)

DESCRIPTOR.message_types_by_name['DeviceEmpty'] = _DEVICEEMPTY
DESCRIPTOR.message_types_by_name['DeviceInfo'] = _DEVICEINFO
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

DeviceEmpty = _reflection.GeneratedProtocolMessageType('DeviceEmpty', (_message.Message,), {
  'DESCRIPTOR' : _DEVICEEMPTY,
  '__module__' : 'device_service_pb2'
  # @@protoc_insertion_point(class_scope:chip.rpc.DeviceEmpty)
  })
_sym_db.RegisterMessage(DeviceEmpty)

DeviceInfo = _reflection.GeneratedProtocolMessageType('DeviceInfo', (_message.Message,), {
  'DESCRIPTOR' : _DEVICEINFO,
  '__module__' : 'device_service_pb2'
  # @@protoc_insertion_point(class_scope:chip.rpc.DeviceInfo)
  })
_sym_db.RegisterMessage(DeviceInfo)



_DEVICE = _descriptor.ServiceDescriptor(
  name='Device',
  full_name='chip.rpc.Device',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=152,
  serialized_end=408,
  methods=[
  _descriptor.MethodDescriptor(
    name='FactoryReset',
    full_name='chip.rpc.Device.FactoryReset',
    index=0,
    containing_service=None,
    input_type=_DEVICEEMPTY,
    output_type=_DEVICEEMPTY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='Reboot',
    full_name='chip.rpc.Device.Reboot',
    index=1,
    containing_service=None,
    input_type=_DEVICEEMPTY,
    output_type=_DEVICEEMPTY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='TriggerOta',
    full_name='chip.rpc.Device.TriggerOta',
    index=2,
    containing_service=None,
    input_type=_DEVICEEMPTY,
    output_type=_DEVICEEMPTY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetDeviceInfo',
    full_name='chip.rpc.Device.GetDeviceInfo',
    index=3,
    containing_service=None,
    input_type=_DEVICEEMPTY,
    output_type=_DEVICEINFO,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_DEVICE)

DESCRIPTOR.services_by_name['Device'] = _DEVICE

# @@protoc_insertion_point(module_scope)
